from flask import Blueprint, request
from flask_jwt_extended import jwt_required
from flask_restful import Api, Resource

from portal.service import service
from portal.tools.decorators import error_handler
from portal.tools.logging import log
from portal.database.models import Group
from portal.rest import group_schema, groups_schema, users_schema, user_list_update_schema, group_import_schema
from portal.service.errors import PortalAPIError, ForbiddenError
from portal.service import policies

groups = Blueprint('groups', __name__)
groups_api = Api(groups)


class GroupResource(Resource):
    @error_handler
    @jwt_required
    def get(self, cid, gid):
        log.info(f"GET to {request.url}")
        client = service.find_client()

        course = service.find_course(cid)
        group = service.find_group(course, gid)
        # authorization
        # note: read_groups should be enabled for students of a course TODO make default?
        if not (policies.check_component(component=client, course=course, permissions=['read_groups'])
                or policies.check_user(user=client, course=course, permissions=['read_groups'])):
            raise ForbiddenError(uid=client.id)
        return group_schema.dump(group)

    @error_handler
    @jwt_required
    def delete(self, cid, gid):
        log.info(f"DELETE to {request.url}")
        client = service.find_client()

        course = service.find_course(cid)
        group = service.find_group(course, gid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['write_groups'])
                or policies.check_user(user=client, course=course, permissions=['write_groups'])):
            raise ForbiddenError(uid=client.id)

        service.delete_entity(group)
        return '', 204

    @error_handler
    @jwt_required
    def put(self, cid, gid):
        log.info(f"PUT to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        group = service.find_group(course, gid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['write_groups'])
                or policies.check_user(user=client, course=course, permissions=['write_groups'])):
            raise ForbiddenError(uid=client.id)

        json_data = request.get_json()
        if not json_data:
            raise PortalAPIError(400, message='No data provided for group update.')

        data = group_schema.load(json_data)[0]

        group.name = data['name']
        service.write_entity(group)
        log.debug(f"Updated group {group.name} to {group}.")
        return '', 204


class GroupsList(Resource):
    @error_handler
    @jwt_required
    def get(self, cid):
        log.info(f"GET to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['read_groups'])
                or policies.check_user(user=client, course=course, permissions=['read_groups'])):
            raise ForbiddenError(uid=client.id)

        return groups_schema.dump(course.groups)

    @error_handler
    @jwt_required
    def post(self, cid):
        log.info(f"POST to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['write_groups'])
                or policies.check_user(user=client, course=course, permissions=['write_groups'])):
            raise ForbiddenError(uid=client.id)

        json_data = request.get_json()
        if not json_data:
            raise PortalAPIError(400, message='No data provided for group creation.')

        data = group_schema.load(json_data)[0]

        new_group = Group(course=course, name=data['name'])
        service.write_entity(new_group)
        log.info(f"Created group={new_group} for course {course.codename}")
        return group_schema.dump(new_group)[0], 201


class GroupUsersList(Resource):
    @error_handler
    @jwt_required
    def get(self, cid, gid):
        log.info(f"GET to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        group = service.find_group(course, gid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['read_groups'])
                or policies.check_user(user=client, course=course, permissions=['read_groups'])):
            raise ForbiddenError(uid=client.id)

        role_id = request.args.get('role')
        if role_id:
            # maybe ineffective - use db query instead
            role = service.find_role(course, role_id)
            return users_schema.dump([user for user in group.users if role in user.roles])
        return users_schema.dump(group.users)

    @error_handler
    @jwt_required
    def put(self, cid, gid):
        log.info(f"PUT to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        group = service.find_group(course, gid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['write_groups'])
                or policies.check_user(user=client, course=course, permissions=['write_groups'])):
            raise ForbiddenError(uid=client.id)

        json_data = request.get_json()
        if not json_data:
            raise PortalAPIError(400, message='No data provided for group user list update.')

        data = user_list_update_schema.load(json_data)[0]

        # everything from users_add is added, THEN everything from users_remove is subtracted
        users_old = set(group.users)
        users_add = data.get('add')
        users_remove = data.get('remove')
        users = set()
        if users_add:
            to_add = service.find_users(data['add'])
            users = users_old.union(to_add)
        if users_remove:
            to_remove = service.find_users(data['remove'])
            users = users_old.difference(to_remove)

        group.users = list(users)
        service.write_entity(group)

        log.info(f"Updated users of group {group.name} to {group.users}.")
        return '', 204


class GroupUser(Resource):
    @error_handler
    @jwt_required
    def put(self, cid, gid, uid):
        log.info(f"PUT to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        group = service.find_group(course, gid)
        user = service.find_user(uid)
        # authorization
        if not(policies.check_component(component=client, course=course, permissions=['write_groups'])
                or policies.check_user(user=client, course=course, permissions=['write_groups'])):
            raise ForbiddenError(uid=client.id)

        if user not in group.users:
            group.users.append(user)
            log.info(f"Added user {user.username} to group {group.name} in course {course.codename}.")
        else:
            log.info(f"User {user.username} is already in group {group.name} in course {course.codename}: no change.")
        return '', 204

    @error_handler
    @jwt_required
    def delete(self, cid, gid, uid):
        log.info(f"DELETE to {request.url}")
        client = service.find_client()
        course = service.find_course(cid)
        group = service.find_group(course, gid)
        user = service.find_user(uid)
        # authorization
        if not (policies.check_component(component=client, course=course, permissions=['write_groups'])
                or policies.check_user(user=client, course=course, permissions=['write_groups'])):
            raise ForbiddenError(uid=client.id)

        try:
            group.users.remove(user)
        except ValueError:
            raise PortalAPIError(400,
                                 message=f"Could not remove user {user.username} from group {group.name} in course "
                                         f"{course.codename}: group does not contain user.")
        log.info(f"Removed user {user.username} from group {group.name} in course {course.codename}.")
        return '', 204


class GroupImport(Resource):
    @error_handler
    @jwt_required
    def put(self, cid):
        log.info(f"DELETE to {request.url}")
        client = service.find_client()
        target_course = service.find_course(cid)
        # authorization
        if not (policies.check_component(component=client, course=target_course, permissions=['write_groups'])
                or policies.check_user(user=client, course=target_course, permissions=['write_groups'])):
            raise ForbiddenError(uid=client.id)

        json_data = request.get_json()
        if not json_data:
            raise PortalAPIError(400, message='No data provided for group import.')

        data = group_import_schema.load(json_data)[0]
        source_course = service.find_course(data['source_course'])
        source_group = service.find_group(source_course, data['source_group'])

        new_group = service.copy_group(source_group, target_course, data['with_users'])
        service.write_entity(new_group)

        log.info(
            f"Copied group {source_group.name} from course {source_course.codename} to course {target_course.codename} "
            f"as {new_group.name}.")
        return group_schema.dump(new_group), 201


groups_api.add_resource(GroupResource, '/<string:gid>')
groups_api.add_resource(GroupsList, '')
groups_api.add_resource(GroupUsersList, '/<string:gid>/users')
groups_api.add_resource(GroupUser, '/<string:gid>/users/<string:uid>')
groups_api.add_resource(GroupImport, '/import')
